(*
 *  TTAPILine component version 2.1
 *  (c)1998 Jury Kornienko
 *  j.kornienko@pf.lv
 *
 *
 *  See for updates in
 *  http://159.148.87.15/pfs/items.asp?s_id=29
 *
 *  TAPI.pas - Translated by Alexander Staubo
 *
 *  Source code provided under some conditions:
 *
 *  - if you use significant parts of the code, please credit me
 *  - do not release a new version of this component without
 *    significant enhancements
 *  - you have to send me all component improvements too
 *
 * Revision History
 * ----------------
 * TAPILine 1.0 - 1.02
 * 11.04.98 - first release
 * 29.04.98 - eventCallInProcess has been handled via Windows functions
 *            beacase of Delphi 2.0 doesn't have TEvent class...
 * 11.05.98 - new method added:
 *            - WriteLargeData (buffer: PChar, WantedToWrite: LONG) -
 *            allows read WantedToWrite data amount to PChar buffer;
 *            - procedure Write corrected;
 *            The folowing "bug" has been fixed (procedures Write
 *            and WriteLargeData ):
 *            sometimes application hangs up. In Delphi debugger GPF
 *            occures in very "harmless" place. For example
 *            count := count + 1;
 *            or something like this.
 *
 * TAPILine 1.4
 * 09.08.98 - (OnNewData event)under society high pressure OnNewData event
 *             has been added. To generate this event TAPILine creates child
 *             thread (single way how to do it). The main event
 *             code is situated in it. In the same time OnNewDataEvent user
 *             procedure will be called from PARENT thread. This event
 *             arrives a cople of times for each Read operation, therefore
 *             application should check this situation. How to use it please
 *             read in readme.txt.
 *             Idea by Mike Speer and other Delphi Developers.
 *          -  procedure WriteLargeData.
 *          -  procedures Read, ReadLargeData and Write has been completely
 *             redesigned again. Read and Write procedures has been implemented
 *             via XXXXXLargeData procedures accordingly.
 *          -  function cbInQue has been added. It returns count bytes in receive
 *             buffer. One Bill Gates company's "bug" (W95 only) has been found:
 *             if line has NOT been opened, but we try to access it via
 *             DirectDeviceAccess this function returns something like 4,500,000 B
 *             count (!?). Of course, may be function ClearCommError is not guilty,
 *             but...
 *          -  GetIcon corrected.
 *          -  new nice .avi for dial dialog.
 *          -  New Class TLogFile.
 *          -  MakeCall function has been corrected.
 *          -  not necessary to restart W'95 if some of TAPI function fail (I just want
 *             to tell that sentence "You are running under W'95 - shutdown is
 *             recomended." has been removed).
 *          -  property CountryCode has been removed
 *
 * 11.08.98 -  The following really serious bug has been fixed
 *             (for all TAPILine versions): sometimes OnConnection event arrives
 *             TWO times (therefore DirectDeviceAccess was performed two
 *             times too). In TAPI documentation is written that this event
 *             can arrives two (or more) times and application should check
 *             may be it already connected. Every time I knew about it, but for
 *             some strange reason in previous component versions I've
 *             implemented this routine for OnConnectionInt+ernal
 *             function only.
 *
 * TAPILine 1.5
 * 09.10.98 -  FDirectAccessScript was dinamically created but never destroyed.
 *             Bug fixed and has been found by Roelof Ridderman.
 *
 * TAPILine 2.0
 * 16.11.98 -  There were many questions like "How do I have to use your
 *             component just for dialing?". Now this feature is implemented
 *             via property CallType. You can choose ctSimpleCall or
 *             ctDataModemCall.
 *          -  ClockLabel - counts connection time. For datamodem it is
 *             quate precisly, but for simple voice call, it doesn't.
 *          -  RingsCount property appears. You can specify this property
 *             if you want that your app takes receive after a while, not
 *             right after call arrived to system.
 *          -  TAPILine has own resource script file now: TL_res.rc
 *             Here are stored all strings constansts, which has been
 *             used in TAPILine. You can make your own string resources.
 *             If you do, please send it to me too.
 *          -  Error strings and other messages are written more precisly
 *             and look better.
 *          -  TAPI callback procedure had been redesigned a little bit.
 *             Now it is more understandable.
 *          -  ReadLargeData and WriteLargeData functions had been
 *             redesigned again. Now they do not "hang up" system during
 *             read/write timeout interval.
 *          -  one "bug" has been fixed in Drop procedure.
 *          -  A few mega functions had been added:
 *             Connect ();
 *             Disconnect ();
 *             to implement "easy data modem app".
 *
 * TAPILine 2.1
 * 26.11.99 -  Thank you for Moshe E. Moskowitz. He sends a TCallerID exploring
 *             routine. I just would like to tell, that CallerID can be known
 *             if your telephone provider supply this information with call (before
 *             first ring). For example latvian LatTelecom doesn't, but
 *             latvian LMT GSM does and so on.
 *
 * TAPILine 2.11
 * 15.04.03 -  Some minor changes of component
 ****)
unit TAPILine;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Tapi, ComCtrls, ExtCtrls, TL_Cnst;

const
  TL_TIMEOUT = 20; // sec.
  HiVer = $00020000;  // Highest API version wanted (2.0)
  LoVer = $00010004;  // Lowest API version accepted (1.4)
  EVENT_TIMEOUT = 300;
  NEW_LINE = Chr (13)+Chr (10);
  CM_NEWDATATHREAD = WM_USER+777;
  CM_NEWDATAARRIVES = WM_USER+778;
  DROP_WAIT_TIMEOUT = 100;
  DROP_WAIT = 10000;

  HEAP_ZERO_MEMORY = $00000008;

type

  TCallType = (ctSimpleCall, ctDataModemCall);

{  TAboutTAPILineProperty = class(TPropertyEditor)
  public
  	procedure Edit; override;
  	function GetAttributes: TPropertyAttributes; override;
  	function GetValue: string; override;
  end;
 }
  TformDialing = class(TForm)
    bHangUp: TButton;
    ModemName: TLabel;
    Image: TImage;
    ClockTimer: TTimer;
    procedure bHangUpClick(Sender: TObject);
    procedure ClockTimerTimer(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
    procedure CMNewData (var Message: TMessage); message CM_NEWDATATHREAD;
    procedure CMNewDataArrives (var Message: TMessage); message CM_NEWDATAARRIVES;
  public
    { Public declarations }
    Animate: TAnimate;
  end;

  TCallerIDEvent = procedure(Sender : TObject; CallerID, CalledID: string) of object;

  TTAPILine = class(TComponent)
  private
    { Private declarations }
//    FAbout: TAboutTAPILineProperty;
    FDirectAccessLogFile: string;{file name}
    FDirectAccessScript: TStrings;
    FDirectAccess: boolean;
    sDirectAccessResult: string;
    FTimeout: DWORD;
    FRingsCount: DWORD;
    FCallType: TCallType;
    FOnCallerID: TCallerIDEvent;

    FOnDisconnect: TNotifyEvent;
    FOnBusy: TNotifyEvent;
    FOnConnection: TNotifyEvent;
    FOnCallAnswering: TNotifyEvent;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FOnDropCall: TNotifyEvent;
    FOnUnexpectedClose: TNotifyEvent;
    FOnNewData: TNotifyEvent;
    FOnCallAccepting: TNotifyEvent;
    FOnMakeCallCompleted: TNotifyEvent;

    FPhoneNumber: string;
    FDevIDToInitialize: DWORD;
    FWaitForCall: boolean;

    eventCallInProcess: THANDLE;
    eventOverlappedWrite: THANDLE;
    eventOverlappedRead: THANDLE;
    eventDropCall: THANDLE;
    eventConnected: THANDLE;
    eventCallArrives: THANDLE;

    FOnCallOffering: TNotifyEvent;{call in}
    bOlreadyDisconnected: boolean;

    dwRequestAnswer: DWORD;
    dwRequestMakeCall: DWORD;
    dwRequestDrop: DWORD;

    timeStarted: TDateTime;
    Procedure OnConnectionInt;
    procedure SetPhoneNumber (value: String);
    procedure SetRingsCount (value: DWORD);
    procedure SetDevIDToInitialize (value: DWORD);
{    function LineCallStateToStr (CallState : DWORD) : string;}
    function LineCallStateInfoToStr (CallState, Info : DWORD) : string;
    function ClearCommErrorToStr (dwErrorFlags : DWORD) : string;
    procedure SetLines (Value: TStrings);
    function GetLineName (lpLineDevCaps_: LPLINEDEVCAPS): string;
    procedure StartClock;

    {TAPI callback messages handler}
    function OnRingingMes (dwParam3: DWORD):string;
    procedure OnConnectionMes;
    function OnCallOfferingMes (hDevice: THANDLE): string;
    procedure OnIdleStateMes;
    procedure OnDisconnectedMes;
    procedure OnBusyMes;
    procedure OnReplyMes (dwParam1: DWORD; dwErrorCode: DWORD);
  public
    { Public declarations }
    Line : HLINE;
    Call : HCALL;
    LineApp : HLINEAPP;

    bDropped: boolean;
    Version : DWORD;
    NumDevs : Integer; {Number of TAPI devices}
    bConnected: boolean;
    InfoLabel: TStatusPanel;
    ClockLabel: TStatusPanel;
    hCommFile: DWORD;
    bStopWaitingNewData: boolean;
    dwLineUsability: DWORD;
    TraceMemo : TMemo;
    function I_lineNegotiateAPIVersion (dwDeviceID: DWORD): DWORD;
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ConfigDialog;
    function Drop: boolean;
    function MakeCall: boolean;
    procedure Open;
    function Write (const sData: String): Boolean;
    function Read(wantedToRead: DWORD):string;
    function ReadLargeData (buffer: PChar; wantedToRead: DWORD):LONG;
    procedure Close;
    procedure TranslateDialog;
    function GetIcon (DevID: DWORD): HICON;
    function FillTAPILine (sl: TStrings): DWORD;
    function VerifyUsableLine (dwDeviceID: DWORD): LONG;
    function WriteLargeData (const buffer: PChar; wantedToWrite: DWORD): Boolean;
    function cbInQue:longint;
    procedure SetTimeout (Value: DWORD);
    procedure ResumeWaitingNewData;
    procedure StopWaitingNewData;
    function Connect(bSynh:boolean):boolean;
    function Disconnect:boolean;
    function IsConnectionWaiting:boolean;
    function IsCallInProcess:boolean;
    function IsConnected:boolean;
    function RetriveCOMHandle:LONG;
    procedure MakeDirectAccessScript;
  protected
    function GenericOverlappedRead (buffer: PChar; wantedToRead: DWORD;
         dwTimeoutMsc: DWORD; hComm: LONG):LONG;
    function GenericOverlappedWrite(const buffer: PChar; wantedToWrite: DWORD;
         dwTimeoutMsc: DWORD; hComm: LONG): Boolean;
    procedure DoCallerID;
  published
//    property About: TAboutTAPILineProperty read FAbout write FAbout;
    property WaitForCall: boolean read FWaitForCall write FWaitForCall;
    property DeviceDirectAccess: boolean read FDirectAccess write FDirectAccess;
    property PhoneNumber: string read FPhoneNumber write SetPhoneNumber;
    property DevIdToInitialize: DWORD read FDevIdToInitialize write SetDevIdToInitialize;
    property DirectAccessScript: TStrings read FDirectAccessScript write SetLines;
    property DirectAccessLogFile: string read FDirectAccessLogFile write FDirectAccessLogFile;
    property Timeout: DWORD read FTimeout write FTimeout;
    property RingsCount: DWORD read FRingsCount write SetRingsCount;
    property CallType: TCallType read FCallType write FCallType;

    property OnConnection: TNotifyEvent read FOnConnection write FOnConnection;
    property OnCallOffering: TNotifyEvent read FOnCallOffering write FOnCallOffering;
    property OnCallAnswering: TNotifyEvent read FOnCallAnswering write FOnCallAnswering;
    property OnDisconnect: TNotifyEvent read FOnDisconnect write FOnDisconnect;
    property OnBusy: TNotifyEvent read FOnBusy write FOnBusy;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnDropCall: TNotifyEvent read FOnDropCall write FOnDropCall;
    property OnUnexpectedClose: TNotifyEvent read FOnUnexpectedClose write FOnUnexpectedClose;
    property OnNewData: TNotifyEvent read FOnNewData write FOnNewData;
    property OnCallAccepting: TNotifyEvent read FOnCallAccepting write FOnCallAccepting;
    property OnMakeCallCompleted: TNotifyEvent read FOnMakeCallCompleted write FOnMakeCallCompleted;
    property OnCallerID: TCallerIDEvent read FOnCallerID write FOnCallerID;
  end;

  TNewData = class(TThread)
  private
    { Private declarations }
  protected
    procedure Execute; override;
  end;

  TLogFile = class(TComponent)
  private
    { Private declarations }
    slLogFile: TStringList;
  public
    { Public declarations }
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;
    procedure WriteLog (tag: string; responce: string);
    procedure SaveToFile (sfile: string);
  published
    { no properties }
  end;

   procedure LineCallbackProc (hDevice,
                            dwMessage,
                            dwInstance,
                            dwParam1,
                            dwParam2,
                            dwParam3 : DWORD); stdcall;

   procedure ZeroMem (p: PChar);
   function LoadString (str: DWORD):string;
   procedure TapiCheck (ResultCode : DWORD);
   procedure Register;

   Var
      thisTAPILine: TTAPILine;
      formDialing: TformDialing;
      thread: TNewData;
      bWaitingInProcess: boolean;
      TAPIUILib : HINST;

implementation
